/*
 * Serial.cpp
 *
 *  Created on: 01 mar 2017
 *      Author: Francesco Gatto
 */

#include <stdio.h>
#ifdef _WIN32
#include <windows.h>
#include <conio.h>
#else
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
#include <sys/ioctl.h>
#endif

#include "Serial.h"

Serial::Serial(const char* portName)
{
#ifdef _WIN32
	hIn = hOut = CreateFileA(portName, GENERIC_READ | GENERIC_WRITE, 0, NULL,
			OPEN_EXISTING, 0, NULL);
	bConsole = bPipe = false;
#else
	// open with non-blocking behavior
	fdOut = fdIn = open(portName, O_RDWR | O_NOCTTY | O_NDELAY);
	// set raw
	struct termios options;
	tcgetattr(fdIn, &options);
//	cfmakeraw(&options);
//	cfsetspeed(&options, B9600);
	options.c_cflag = B9600 | CS8 | CLOCAL | CREAD;
	options.c_iflag = IGNPAR;
	options.c_oflag = 0;
	options.c_lflag = 0;
	tcsetattr(fdIn, TCSAFLUSH, &options);
#endif
}

#ifdef _WIN32
Serial::Serial(HANDLE in, HANDLE out)
{
	hIn = in;
	hOut = out;
	bConsole = bPipe = false;
}
#else
Serial::Serial(int in, int out)
{
  fdIn = in;
  fdOut = out;
  // set non-blocking read
  int flags = fcntl(fdIn, F_GETFL, 0);
  fcntl(fdIn, F_SETFL, flags | O_NONBLOCK);
}
#endif

Serial::~Serial()
{
#ifdef _WIN32
	CloseHandle(hIn);
	if (hOut != hIn)
		CloseHandle(hOut);
#else
	close(fdIn);
#endif
}

void Serial::begin(unsigned long baud)
{
#ifdef _WIN32
	// check device handle type
	DWORD n = MAXDWORD;
	if (GetConsoleMode(hIn, &n))
	{
		bConsole = true;
		_kbhit();
		n = ENABLE_ECHO_INPUT | ENABLE_LINE_INPUT;
		SetConsoleMode(hIn, n);
		FlushConsoleInputBuffer(hIn);
		_kbhit();
	}
	else if (GetFileType(hIn) == FILE_TYPE_PIPE)
	{
		bPipe = true;
		PeekNamedPipe(hIn, NULL, 0, NULL, &n, NULL);
		n = PIPE_READMODE_BYTE | PIPE_NOWAIT;
		SetNamedPipeHandleState(hIn, &n, NULL, NULL);
		PeekNamedPipe(hIn, NULL, 0, NULL, &n, NULL);
	}
	else // assume COM port
	{
	  SetCommMask(hIn, EV_BREAK);
		// set non-blocking read
		COMMTIMEOUTS cto;
		cto.ReadIntervalTimeout = MAXDWORD;
		cto.ReadTotalTimeoutConstant = 0;
		cto.ReadTotalTimeoutMultiplier = 0;
		cto.WriteTotalTimeoutConstant = 0;
		cto.WriteTotalTimeoutMultiplier = 0;
		SetCommTimeouts(hIn, &cto);
		// set baudrate and handshake (need RTS low for QuickUSB)
		DCB dcb;
		dcb.DCBlength = sizeof(dcb);
		GetCommState(hIn, &dcb);
		dcb.BaudRate = DWORD(baud);
		dcb.Parity = NOPARITY;
		dcb.ByteSize = 8;
		dcb.StopBits = ONESTOPBIT;
		dcb.fOutxCtsFlow = FALSE;
    dcb.fOutxDsrFlow = FALSE;
    dcb.fOutX = FALSE;
    dcb.fInX = FALSE;
		dcb.fRtsControl = RTS_CONTROL_DISABLE;
    dcb.fDtrControl = DTR_CONTROL_DISABLE;
		SetCommState(hIn, &dcb);
    Sleep(50);
		EscapeCommFunction(hIn, SETRTS);
    EscapeCommFunction(hIn, SETDTR);
    // clear buffers
    Sleep(50);
    PurgeComm(hIn, PURGE_TXCLEAR | PURGE_RXCLEAR);
	}
#else
	struct termios options;
	tcgetattr(fdIn, &options);
	speed_t speed = B0;
	switch (baud)
	{
	case 9600:
		speed = B9600;
		break;
	case 115200:
		speed = B115200;
		break;
	}
	cfsetspeed(&options, speed);
	tcsetattr(fdIn, TCSANOW, &options);
	// set handshake (for QuickUSB)
	int status = 0;
	ioctl(fdIn, TIOCMGET, &status);
	status |= TIOCM_RTS;
	ioctl(fdIn, TIOCMSET, &status);
#endif
}

void Serial::end()
{

}

int Serial::read()
{
	uint8_t rx = 0;
  if (available() < 1)
    return -1;
#ifdef _WIN32
	DWORD n = 0;
	if (bConsole)
		rx = _getch(), n = 1;
	else
		ReadFile(hIn, &rx, 1, &n, NULL);
	if (n == 1)
		return rx;
#else
	if (fdIn != -1)
	{
    int rx_length = ::read(fdIn, &rx, 1);
    if (rx_length == 1)
      return rx;
	}
#endif
	return -1;
}

size_t Serial::write(uint8_t tx)
{
#ifdef _WIN32
	DWORD n = 0;
	WriteFile(hOut, &tx, 1, &n, NULL);
	if (n != 1)
		return 0;
#else
	if (fdOut != -1)
	{
		int count = ::write(fdOut, &tx, 1);
		if (count != 1)
		  return 0;
	}
#endif
	return 1;
}

void Serial::flush()
{
#ifdef _WIN32
	FlushFileBuffers(hOut);
#else
	tcflush(fdIn, TCIFLUSH);
	int n = 0;
	do {
	  ioctl(fdOut, TIOCOUTQ, &n);
	}
	while (n > 0);
#endif
}

int Serial::available()
{
	int bytesAv = 0;
#ifdef _WIN32
	DWORD n = 0;
	if (bConsole)
	{
		return _kbhit() ? 1 : 0;
	}
	else if (bPipe)
	{
		PeekNamedPipe(hIn, NULL, 0, NULL, &n, NULL);
		return n;
	}
	else
	{
		DWORD errors = 0;
		COMSTAT csta;
		ZeroMemory(&csta, sizeof(csta));
		if (ClearCommError(hIn, &errors, &csta))
			bytesAv = csta.cbInQue;
		else
		{
			errors = WaitForSingleObject(hIn, 0);
			if (errors == WAIT_OBJECT_0)
				bytesAv = 1;
		}
	}
#else
	if (ioctl(fdIn, FIONREAD, &bytesAv) < 0)
	{
		return 0;
	}
	if (bytesAv < 1)
	{
		return 0;
	}
#endif
	return bytesAv;
}
